import math
from copy import deepcopy

from magnumapi.geometry.Geometry import Geometry
from magnumapi.geometry.SlottedGeometry import SlottedRelativeCosThetaGeometry
from magnumapi.geometry.blocks.Block import Block
from magnumapi.geometry.blocks.CosThetaBlock import AbsoluteCosThetaBlock
from magnumapi.geometry.primitives.Line import Line
from pymbse.optim.design_variable.Individual import Individual


class GeometryChange:

    @staticmethod
    def update_global_variables(geometry, individual: Individual) -> Geometry:
        geometry = deepcopy(geometry)
        for global_dv in individual.get_global_dvs():
            if hasattr(geometry, global_dv.variable):
                setattr(geometry, global_dv.variable, global_dv.value)

        return geometry

    @classmethod
    def update_layer_variables(cls, geometry, individual: Individual) -> Geometry:
        geometry = deepcopy(geometry)
        for layer_dv in individual.get_layer_dvs():
            layer_def = geometry.layer_defs[layer_dv.layer - 1]
            if hasattr(layer_def, layer_dv.variable):
                setattr(layer_def, layer_dv.variable, layer_dv.value)
        return geometry

    @classmethod
    def update_block_variables(cls, geometry, individual: Individual) -> Geometry:
        geometry = deepcopy(geometry)
        for block_dv in individual.get_block_dvs():
            cls.update_block_variable(geometry, block_dv, block_dv.bcs)
        return geometry

    @classmethod
    def update_block_variable(cls, geometry, dv, block):
        block_index = geometry.layer_defs[dv.layer - 1].blocks[block - 1]
        index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
        block_def = geometry.blocks[index_in_blocks].block_def
        if hasattr(block_def, dv.variable):
            setattr(block_def, dv.variable, dv.value)

    @classmethod
    def update_multiblock_variables(cls, geometry, individual: Individual) -> Geometry:
        geometry = deepcopy(geometry)
        for multiblock_dv in individual.get_multiblock_dvs():
            for index in range(multiblock_dv.block_start, multiblock_dv.block_end):
                cls.update_block_variable(geometry, multiblock_dv, index)
        return geometry

    @classmethod
    def limit_minimum_phi_r(cls, geometry, min_length=1) -> Geometry:
        geometry = deepcopy(geometry)
        for layer_def in geometry.layer_defs:
            for blocks_index, block_index in enumerate(layer_def.blocks):
                if blocks_index > 0:
                    index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                    radius = geometry.blocks[index_in_blocks].block_def.radius
                    phi_r = geometry.blocks[index_in_blocks].block_def.phi_r
                    if radius > 0:
                        geometry.blocks[index_in_blocks].block_def.phi_r = max(phi_r, math.degrees(min_length / radius))

        return geometry

    @classmethod
    def update_nco_varying_blocks(cls,
                                  geometry,
                                  individual: Individual) -> SlottedRelativeCosThetaGeometry:
        geometry = deepcopy(geometry)

        nco_dvs = [dv for dv in individual.get_block_dvs() if dv.variable == 'nco']
        block_index = 1
        blocks = []
        # for each layer
        for layer_index, layer_def in enumerate(geometry.layer_defs, start=1):
            # find a reference block - first block in layer_def
            block_layer_index_ref = layer_def.blocks[0]
            block_index_ref = geometry.get_index_in_blocks_for_layer_block_index(block_layer_index_ref)
            block_def_ref = geometry.blocks[block_index_ref]

            # find all ncos that belong to that layer and are non-zero
            # update a copy of the reference block as long as there are non-zero blocks - todo: sort the keys
            layer_blocks = []

            nco_layer_dvs = [dv for dv in nco_dvs if (dv.value > 0) and (dv.layer == layer_index)]
            for nco_layer_dv in nco_layer_dvs:
                block_def_ref = deepcopy(block_def_ref)
                block_def_ref.block_def.nco = nco_layer_dv.value
                block_def_ref.block_def.no = block_index
                layer_blocks.append(block_index)
                block_index += 1
                blocks.append(block_def_ref)

            layer_def.blocks = layer_blocks

        return SlottedRelativeCosThetaGeometry(r_aperture=geometry.r_aperture,
                                               blocks=blocks,
                                               layer_defs=geometry.layer_defs)

    @classmethod
    def update_alpha_radial(cls, geometry, individual: Individual):
        geometry = deepcopy(geometry)
        alpha_r_dvs = [block_dv for block_dv in individual.get_block_dvs() if block_dv.variable == 'alpha_rad_r']
        for alpha_r_dv in alpha_r_dvs:

            # block index
            if alpha_r_dv.bcs <= len(geometry.layer_defs[alpha_r_dv.layer - 1].blocks):
                no_block = geometry.layer_defs[alpha_r_dv.layer - 1].blocks[alpha_r_dv.bcs - 1]
                block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
                geometry.blocks[block_index].block_def.phi_r += alpha_r_dv.value

        return geometry

    @classmethod
    def update_phi_r(cls, geometry: Geometry, individual: Individual) -> Geometry:
        geometry = deepcopy(geometry)
        phi_r_dvs = [block_dv for block_dv in individual.get_block_dvs() if block_dv.variable == 'phi_r']
        for phi_r_dv in phi_r_dvs:

            # block index
            if phi_r_dv.bcs <= len(geometry.layer_defs[phi_r_dv.layer - 1].blocks):
                no_block = geometry.layer_defs[phi_r_dv.layer - 1].blocks[phi_r_dv.bcs - 1]
                block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
                geometry.blocks[block_index].block_def.phi_r = phi_r_dv.value

        return geometry

    @classmethod
    def calculate_radial_alpha(cls, geometry: Geometry) -> Geometry:
        """ Method correcting radiality of cos-theta blocks in a geometry.

        """
        # check whether areas were initialize before performing radiality correction
        is_any_area_initialized = any([area for block in geometry.blocks for area in block.areas])

        # Set alpha equal to phi for all blocks except the mid-plane
        geometry = cls._set_alpha_equal_to_phi(geometry)

        # Build blocks and calculate turn positions
        geometry.build_blocks()

        # Calculate alpha corrections
        alpha_corrections = cls._calculate_alpha_corrections(geometry)

        # Correct alpha
        geometry = cls._apply_alpha_correction(geometry, alpha_corrections)

        # Return to the external context of the geometry before entering
        if is_any_area_initialized:
            geometry.build_blocks()
        else:
            geometry.empty_block_areas()

        return geometry

    @staticmethod
    def _set_alpha_equal_to_phi(geometry):
        for layer_def in geometry.layer_defs:
            for block_index in layer_def.blocks[1:]:
                index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                block = geometry.blocks[index_in_blocks]
                if isinstance(block, AbsoluteCosThetaBlock):
                    block.block_def.alpha = block.block_def.phi
        return geometry

    @classmethod
    def _calculate_alpha_corrections(cls, geometry):
        alpha_corrections = []
        for layer_def in geometry.layer_defs:
            for index, block_index in enumerate(layer_def.blocks):
                index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                block = geometry.blocks[index_in_blocks]
                # Calculate correction for all blocks except the mid-plane one
                if index and isinstance(block, AbsoluteCosThetaBlock):
                    alpha_correction = cls.compute_radial_alpha(block)
                else:
                    alpha_correction = 0.0
                alpha_corrections.append(alpha_correction)

        return alpha_corrections

    @staticmethod
    def compute_radial_alpha(block: Block) -> float:
        """ Static method calculating a radiality correction angle for a cos-theta block. It is based on an assumption
        that block alpha angle (inclination angle) is equal to phi angle (positioning angle).

        :param block: a block instance with initialized areas
        :return: an alpha angle correction factor
        """
        if not block.areas:
            raise IndexError('The list of areas is empty, please build block first with build_block() method.')

        l_down = block.areas[0].get_line(0)
        p_down = l_down.p1
        l_up = block.areas[-1].get_line(2)
        p_up = l_up.p2

        angle_ref = (p_up.get_phi() + p_down.get_phi()) / 2

        alpha_down = angle_ref - Line.calculate_relative_alpha_angle(l_down)
        alpha_up = Line.calculate_relative_alpha_angle(l_up) - angle_ref

        return (alpha_down - alpha_up) / 2

    @staticmethod
    def _apply_alpha_correction(geometry, alpha_corrections):
        for index, block in enumerate(geometry.blocks):
            if isinstance(block, AbsoluteCosThetaBlock):
                block.block_def.alpha += alpha_corrections[index]

        return geometry

    @classmethod
    def update_layer_indexing(cls, geometry) -> Geometry:
        """ Method updating block and layer indexing in case a block has no turns.
        The algorithm works as follows:
        1. Take old block indices and map them to None
        2. Keep blocks with at least one conductor
        3. Update the numbering of blocks; update mapping from old to new indices from point 1
        4. Update block indices in each layer definition with map; if a new index is None, then it means that a block
        was removed

        :return: method returns a dictionary from old to new block indices
        """
        # # initialize map
        geometry = deepcopy(geometry)
        old_to_new_index = {block.block_def.no: None for block in geometry.blocks}

        # # remove the missing blocks
        # # keep only block indices that are not missing
        blocks = []
        block_index = 1
        for index, block in enumerate(geometry.blocks):
            if block.block_def.nco > 0:
                block.block_def.no = block_index
                blocks.append(block)
                old_to_new_index[block.block_def.no] = block_index
                block_index += 1

        geometry.blocks = blocks

        # # update layers definition
        new_layer_defs = []
        layer_index = 1
        for layer_def in geometry.layer_defs:
            updated_blocks = [old_to_new_index[block] for block in layer_def.blocks]
            updated_blocks = [updated_block for updated_block in updated_blocks if updated_block is not None]
            if updated_blocks:
                new_layer_defs.append(layer_def)
                layer_def.blocks = updated_blocks
                layer_def.no = layer_index
                layer_index += 1

        geometry.layer_defs = new_layer_defs
        return geometry

    @classmethod
    def update_nco_r(cls, geometry, individual: Individual):
        block_dvs = individual.get_block_dvs()
        nco_r_dvs = [block_dv for block_dv in block_dvs if block_dv.variable == 'nco_r']
        geometry = deepcopy(geometry)

        for nco_r_dv in nco_r_dvs:
            no_block = geometry.layer_defs[nco_r_dv.layer - 1].blocks[nco_r_dv.bcs - 1]
            block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
            block_def_curr = geometry.blocks[block_index].block_def
            block_def_next = geometry.blocks[block_index + 1].block_def
            if nco_r_dv.value >= 0:
                block_def_curr.nco += min(nco_r_dv.value, block_def_next.nco)
                block_def_next.nco -= min(nco_r_dv.value, block_def_next.nco)
            else:
                block_def_curr.nco -= min(abs(nco_r_dv.value), block_def_curr.nco)
                block_def_next.nco += min(abs(nco_r_dv.value), block_def_next.nco)

        geometry = GeometryChange.update_layer_indexing(geometry)
        return geometry



import numpy as np


def calc_rho_cu_nist(T, B, RRR):

    if not isinstance(T, np.ndarray):
        T = np.array([T])

    p1 = 1.171e-17
    p2 = 4.49
    p3 = 3.841e10
    p4 = 1.14
    p5 = 50
    p6 = 6.428
    p7 = 0.4531

    B = abs(B)

    rho_0 = 1.553e-8 / RRR
    rho_i = p1 * np.power(T, p2) / (1 + p1 * p3 * np.power(T, p2 - p4)) * np.exp(-np.power(p5 / T, p6))
    rho_i0 = p7 * rho_i * rho_0 / (rho_i + rho_0)

    rho_n = rho_0 + rho_i + rho_i0

    if B > 0.01:
        a0 = -2.662
        a1 = 0.3168
        a2 = 0.6229
        a3 = -0.1839
        a4 = 0.01827

        x = 1.553e-8 * B / rho_n

        log_x = np.log10(x)
        f_exp = a0 + a1 * log_x + a2 * np.power(log_x, 2) + a3 * np.power(log_x, 3) + a4 * np.power(log_x, 4)
        corr = np.power(10, f_exp)
    else:
        corr = 0

    return rho_n * (1 + corr)



import os
from subprocess import Popen, PIPE

import pandas as pd
from ansys.mapdl.core import Mapdl, launch_mapdl

from magnumapi.tool_adapters.ToolAdapter import ToolAdapter
import pymbse.commons.text_file as text_file


class AnsysToolAdapter(ToolAdapter):
    """ An AnsysToolAdapter class with methods to execute an ANSYS model from command line, read figures of merit table
    and prepare a Lorentz force file

    """

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 model_file: str,
                 rst_file: str
                 ) -> None:
        """ A constructor of AnsysToolAdapter instance

        :param input_folder_rel_dir: a relative directory with model inputs
        :param input_file: a name of an input file
        :param output_file: a name of an output file
        :param model_file: a name of a model file
        :param rst_file: a name of an output rst file
        """
        self.root_dir = os.getcwd()
        self.input_folder_rel_dir = input_folder_rel_dir
        self.input_file = input_file
        self.output_file = output_file
        self.model_file = model_file
        self.rst_file = rst_file
        self.output_lines = []

    def get_input_path(self):
        return os.path.join(self.root_dir, self.input_folder_rel_dir, self.input_file)

    def get_output_path(self):
        return os.path.join(self.root_dir, self.input_folder_rel_dir, self.output_file)

    def get_model_path(self):
        return os.path.join(self.root_dir, self.input_folder_rel_dir, self.model_file)

    def run(self) -> None:
        raise NotImplementedError('This method is not implemented for this class')

    def read_figures_of_merit_table(self) -> pd.DataFrame:
        output_path = os.path.join(self.root_dir, self.input_folder_rel_dir, self.output_file)
        return self.read_figures_of_merit_table_from_file(output_path)

    @staticmethod
    def read_figures_of_merit_table_from_file(output_path):
        with open(output_path, 'r') as output:
            output_file_content = output.read().split('\n')
        keys = output_file_content[0].split(' ')
        keys = [key for key in keys if key != '']
        values = output_file_content[1].split(' ')
        values = [float(value) for value in values if value != '']
        key_value = {key: value for key, value in zip(keys, values)}
        return pd.DataFrame(key_value, index=[0])

    @staticmethod
    def convert_figures_of_merit_to_dict(fom_df: pd.DataFrame) -> dict:
        return fom_df.to_dict('records')[0]

    def prepare_force_file(self, input_force_file: str, output_force_file: str) -> None:
        """ Method preparing a force file for ANSYS from a ROXIE Lorentz force file, .force2d

        :param input_force_file: a name of a ROXIE Lorentz force file
        :param output_force_file: a name of an output file for ANSYS
        """
        input_force_file_path = os.path.join(self.root_dir, self.input_folder_rel_dir, input_force_file)
        output_force_file_path = os.path.join(self.root_dir, self.input_folder_rel_dir, output_force_file)

        AnsysToolAdapter.convert_roxie_force_file_to_ansys(input_force_file_path, output_force_file_path)

    @staticmethod
    def convert_roxie_force_file_to_ansys(input_force_file_path: str, output_force_file_path: str) -> None:
        """ Static method preparing a force file for ANSYS from a ROXIE Lorentz force file, .force2d

        :param input_force_file: a name of a ROXIE Lorentz force file
        :param output_force_file: a name of an output file for ANSYS
        """
        force_txt = text_file.readlines(input_force_file_path)
        
        ansys_force = []
        for force_txt_el in force_txt:
            row_float = [float(el) for el in force_txt_el.replace('\n', '').split(' ') if el != '']
            ansys_force.append('nodeNum = NODE(%f, %f, 0.0)' % tuple(row_float[:2]))
            ansys_force.append('F,nodeNum,FX, %f' % row_float[2])
            ansys_force.append('F,nodeNum,FY, %f' % row_float[3])

        text_file.writelines(output_force_file_path, ansys_force)

    @staticmethod
    def remove_ansys_output_files(root_dir: str, n_outputs: int):
        for i in range(n_outputs):
            ansys_output_file = 'vallone_%d.out' % i
            ansys_output_path = os.path.join(root_dir, ansys_output_file)
            if os.path.isfile(ansys_output_path):
                os.remove(ansys_output_path)

    @staticmethod
    def read_multiple_ansys_figures_of_merit(root_dir: str, n_outputs: int) -> list:
        fom_dcts = []

        for i in range(n_outputs):
            ansys_output_file = 'vallone_%d.out' % i
            ansys_output_path = os.path.join(root_dir, ansys_output_file)
            if os.path.isfile(ansys_output_path) and os.path.getsize(ansys_output_path) > 0:
                fom_df = AnsysToolAdapter.read_figures_of_merit_table_from_file(ansys_output_path)
                if fom_df['seqv'].values[0] < 1e-6:
                    fom_dct = {"seqv": float('nan')}
                else:
                    fom_dct = AnsysToolAdapter.convert_figures_of_merit_to_dict(fom_df)
            else:
                fom_dct = {"seqv": float('nan')}

            fom_dcts.append(fom_dct)

        return fom_dcts


class TerminalAnsysToolAdapter(AnsysToolAdapter):

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 model_file: str,
                 rst_file: str,
                 out_temp_file: str,
                 executable_path: str,
                 license_type: str) -> None:
        """ A constructor of TerminalAnsysToolAdapter instance

        :param input_folder_rel_dir: a relative directory with model inputs
        :param input_file: a name of an input file
        :param output_file: a name of an output file
        :param model_file: a name of a model file
        :param rst_file: a name of an output rst file
        :param out_temp_file: a path to an output file
        :param executable_path: an absolute path to an ANSYS APDL executable
        :param license_type: a license type for ANSYS
        """
        super().__init__(input_folder_rel_dir, input_file, output_file, model_file, rst_file)
        self.input_file = input_file
        self.output_file = output_file
        self.model_file = model_file
        self.out_temp_file = out_temp_file
        self.executable_path = executable_path
        self.license_type = license_type

    def run(self):
        output_path = os.path.join(self.root_dir, self.input_folder_rel_dir, self.output_file)
        if os.path.isfile(output_path):
            os.remove(output_path)

        model_dir = os.path.join(self.root_dir, self.input_folder_rel_dir)
        input_path = os.path.join(self.root_dir, self.input_folder_rel_dir, self.input_file)
        out_temp_path = os.path.join(self.root_dir, self.input_folder_rel_dir, self.out_temp_file)
        cmd = [r'%s' % self.executable_path, '-b', '-p', self.license_type, '-smp', '-np', '2', '-lch',
               '-dir', model_dir, '-i', input_path, '-o', out_temp_path]
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        byte_output, err = p.communicate(b"input data that is passed to subprocess' stdin")

        byte_output_decode = byte_output.decode()
        self.output_lines = byte_output_decode.split('\n')


class MapdlAnsysToolAdapter(AnsysToolAdapter):

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 model_file: str,
                 rst_file: str,
                 single_upload_files: list,
                 reupload_files: list) -> None:
        """

        :param input_folder_rel_dir:
        :param input_file:
        :param output_file:
        :param model_file:
        :param rst_file: a name of an output rst file
        :param single_upload_files:
        :param reupload_files:
        """
        super().__init__(input_folder_rel_dir, input_file, output_file, model_file, rst_file)
        self.single_upload_files = single_upload_files
        self.reupload_files = reupload_files
        self.mapdl = None

    def _launch_mapdl(self):
        self.mapdl = launch_mapdl()

    def run(self):

        if not self.mapdl:
            self._launch_mapdl()
            for single_upload_file in self.single_upload_files:
                self.mapdl.upload(os.path.join(self.root_dir, self.input_folder_rel_dir, single_upload_file))

        for reupload_file in self.reupload_files:
            self.mapdl.upload(os.path.join(self.root_dir, self.input_folder_rel_dir, reupload_file))

        self.mapdl.input(self.input_file)
        self.mapdl.finish()

        self.mapdl.download(self.output_file, os.path.join(self.root_dir, self.input_folder_rel_dir, self.output_file))
        self.mapdl.download(self.rst_file, os.path.join(self.root_dir, self.input_folder_rel_dir, self.rst_file))


class DockerMapdlAnsysToolAdapter(MapdlAnsysToolAdapter):

    def __init__(self,
                 input_folder_rel_dir: str,
                 input_file: str,
                 output_file: str,
                 model_file: str,
                 rst_file: str,
                 single_upload_files: list,
                 reupload_files: list,
                 ip: str,
                 port: int) -> None:
        """

        :param input_folder_rel_dir:
        :param input_file:
        :param output_file:
        :param model_file:
        :param rst_file:
        :param single_upload_files:
        :param reupload_files:
        :param ip:
        :param port:
        """
        super().__init__(input_folder_rel_dir=input_folder_rel_dir,
                         input_file=input_file,
                         output_file=output_file,
                         model_file=model_file,
                         rst_file=rst_file,
                         single_upload_files=single_upload_files,
                         reupload_files=reupload_files)
        self.ip = ip
        self.port = port

    def _launch_mapdl(self):
        self.mapdl = Mapdl(ip=self.ip, port=self.port, request_instance=False)



import math
from dataclasses import dataclass


@dataclass(frozen=True)
class Point:
    """A Point class implementing basic operations of a cartesian point in 2D.

   Attributes:
       x (float): x coordinate
       y (float): y coordinate
    """
    x: float
    y: float

    # ToDo add examples to documentation and test them

    @staticmethod
    def of_cartesian(x: float, y: float) -> "Point":
        """ Static method creating a new Point object with cartesian x and y coordinates

        :param x: x coordinate
        :param y: y coordinate
        :return: a new Point object with x,y coordinates
        """
        return Point(x=x, y=y)

    @staticmethod
    def of_polar(r: float, phi: float) -> "Point":
        """ Static method creating a new Point object with polar r and phi coordinates.

        :param r: point radius
        :param phi: point angle (in degrees)
        :return: a new Point object with r, phi coordinates
        """
        phi_in_rad = math.radians(phi)
        x = r * math.cos(phi_in_rad)
        y = r * math.sin(phi_in_rad)
        return Point(x=x, y=y)

    def copy(self) -> "Point":
        """ Method returning a copy of a current Point instance

        :return: a copy of the Point instance
        """
        return Point.of_cartesian(x=self.x, y=self.y)

    def get_phi(self) -> float:
        """ Method returning point angle in degrees

        :return: point angle in polar coordinate expressed in degrees
        """
        return math.degrees(math.atan2(self.y, self.x))

    def get_phi_in_rad(self) -> float:
        """ Method returning point angle in radians

        :return: point angle in polar coordinate expressed in radians
        """
        return math.atan2(self.y, self.x)

    def get_r(self) -> float:
        """ Method returning point radius

        :return: point radius in polar coordinates
        """
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def rotate(self, alpha: float) -> "Point":
        """ Method rotating a Point instance by a given angle w.r.t. the center of the coordinate system (0,0)

        :param alpha: rotation angle in degrees
        :return: a new rotated Point instance
        """
        alpha_in_rad = math.radians(alpha)
        x = math.cos(alpha_in_rad) * self.x - math.sin(alpha_in_rad) * self.y
        y = math.sin(alpha_in_rad) * self.x + math.cos(alpha_in_rad) * self.y
        return Point.of_cartesian(x=x, y=y)

    def translate(self, point: "Point") -> "Point":
        """ Method translating a Point instance by an input point

        :param point: input point by which a Point instance is translated
        :return: a new translated Point instance
        """
        # ToDo this method can profit from add method
        return Point.of_cartesian(x=self.x + point.x, y=self.y + point.y)

    def plot(self, ax, marker='o', markersize=3, color="black") -> None:
        """ Method plotting point on an input axis.

        :param ax: axis on which a point is plotted
        :param marker: point marker, circle by default
        :param markersize: point marker size, 3 by default
        :param color: point marker color, black by default
        :return: None
        """
        ax.plot([self.x], [self.y], marker=marker, markersize=markersize, color=color)

    def __neg__(self) -> "Point":
        """ Method negating point coordinates

        :return: a new a Point instance with negative x and y coordinates
        """
        return Point.of_cartesian(-self.x, -self.y)

    def __sub__(self, p: "Point") -> "Point":
        """ Method subtracting a Point p from a Point instance

        :param p: Point object subtracted from a Point instance
        :return: a new Point instance equal to a difference between a Point instance and input point
        """
        return Point.of_cartesian(self.x - p.x, self.y - p.y)

    def __add__(self, p: "Point") -> "Point":
        """ Method adding a Point p to a Point instance

        :param p: Point object added to a Point instance
        :return: a new Point instance equal to a sum of a Point instance and input point
        """
        return Point.of_cartesian(self.x + p.x, self.y + p.y)

    def __mul__(self, scalar: float) -> "Point":
        """ Method multiplying a Point p by a Point instance

        :param p: Point object multiplied by a Point instance
        :return: a new Point instance equal to a multiplication of a Point instance and input point
        """
        # ToDo lmul, rmul missing
        return Point.of_cartesian(self.x * scalar, self.y * scalar)

    def __str__(self) -> str:
        """ Method returning a string representation of a Point for a user

        :return: string representation of a Point instance
        """
        if isinstance(self.x, int) and isinstance(self.y, int):
            return "(%d, %d)" % (self.x, self.y)
        elif isinstance(self.x, int):
            return "(%d, %f)" % (self.x, self.y)
        elif isinstance(self.y, int):
            return "(%f, %d)" % (self.x, self.y)
        else:
            return "(%f, %f)" % (self.x, self.y)

    def __repr__(self) -> str:
        """ Method returning a string representation of a Point for a developer

        :return: string representation of a Point instance as for a constructor call
        """
        return "Point(%f, %f)" % (self.x, self.y)

    def __eq__(self, p: "Point") -> bool:
        """ Method checking equality between two points

        :param p: a Point instance with which equality is checked
        :return: True if points are exactly equal (no tolerance is used), False otherwise
        """
        return self.x == p.x and \
               self.y == p.y

    def det(self) -> float:
        """ Method calculating length of a vector constructed from the origin of the cartesian coordinate system (0, 0)
        and a Point instance coordinates. This method is an alias for Point().get_r() method.

        :return: length of a vector constructed from the coordinate system origin and the point itself
        """
        return self.get_r()

    def dot(self, p: "Point") -> float:
        """ Method calculating the dot product of a Point instance with input Point p

        :param p: input Point with which the dot product is calculated
        :return: scalar value of the dot product
        """
        return self.x * p.x + self.y * p.y

    def unit(self) -> "Point":
        """ Method calculating a unit vector of a vector constructed from the origin of the cartesian coordinate system
        (0, 0) and a Point instance coordinates.

        :return: a Point instance with coordinates scaled with the length of the length in order to obtain a unit length
        """
        det = self.det()
        if det == 0:
            return Point.of_cartesian(0.0, 0.0)
        else:
            return Point.of_cartesian(self.x / det, self.y / det)



from typing import Tuple, Union
from dataclasses import dataclass

import pandas as pd
from shapely.geometry import Polygon
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Arc import Arc
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Point import Point


@dataclass(frozen=True)
class Area:
    """ An Area class implementing a 2D area made of connected lines/arcs

   Attributes:
       lines (Tuple): tuple of Line/Arc objects
    """
    lines: Tuple[Union[Line, Arc], ...]

    @staticmethod
    def of_lines(lines: Tuple[Union[Line, Arc], ...]) -> "Area":
        """ Static method creating an Area object from a tuple of Line objects. Line objects are copied to preserve
        static behaviour and avoid side effects.

        :param lines: tuple of Line objects
        :return: An area with copy of input Line objects
        """
        return Area(lines=tuple(line.copy() for line in lines))

    @staticmethod
    def of_closed_ordered_list_of_points(points: Tuple[Point, ...]) -> "Area":
        #ToDo: add test
        #ToDo: add docstring
        lines = []
        for ix in range(len(points) - 1):
            lines.append(Line.of_end_points(points[ix], points[ix + 1]))

        lines.append(Line.of_end_points(points[-1], points[0]))

        return Area.of_lines(tuple(lines))

    def get_line(self, index: int) -> Line:
        """ Method returning a line at a given index.

        :param index: index of line to be returned
        :return: Line at the given index
        """
        if abs(index) >= len(self.lines):
            raise IndexError('The requested line index %d is out of range for a list of length %d' %
                             (index, len(self.lines)))
        return self.lines[index].copy()

    def copy(self) -> "Area":
        """ Method returning a copy of an Area instance

        :return: a new Area instance with copy of lines
        """
        return Area(lines=tuple(line.copy() for line in self.lines))

    def translate(self, vector: Point) -> "Area":
        """ Method translating an Area instance by a given vector

        :param vector: vector by which an area is translated
        :return: a translated area
        """
        return Area(lines=tuple(line.translate(vector) for line in self.lines))

    def rotate(self, angle: float) -> "Area":
        """ Method rotating area by a given angle w.r.t. origin of the coordinate system (0, 0)

        :param angle: rotation angle in degrees
        :return: rotated area
        """
        return Area(lines=tuple(line.rotate(angle) for line in self.lines))

    def plot(self, ax: plt.Axes) -> None:
        """ Method plotting an Area instance on an input axis

        :param ax: input axis for plot
        """
        for line in self.lines:
            line.plot(ax)

    def __str__(self):
        """ Method returning a string representation of an Area for a user

        :return: string representation of an Area instance
        """
        return ",".join([str(line) for line in self.lines])

    def to_df(self):
        """ Method converting a line object into a dataframe with end point coordinates of each line. Lines are
        distinguished by 1-based index.

        :return: a DataFrame with x1_i, y1_i, x2_i, y2_i columns, where i is the line index (1-based)
        """
        return pd.concat([line.to_df().add_suffix('_%d' % (i + 1)) for i, line in enumerate(self.lines)], axis=1)

    @staticmethod
    def find_trapezoid_shift_to_intercept_with_radius(radius: float, area: "Area", eps=1e-6) -> float:
        """ Method finding a shift for a trapezoidal area to get in contact with a radius.
        The method works as follows:

        - for lower (index 0) line of an area, find an intersection point with a radius
        - for upper (index 2) line of an area, find an intersection point with a radius

        - find distance between the intersection point and lower line end point closest to the radius (p2)
        - find distance between the intersection point and upper line end point closest to the radius (p1)

        - take the shortest distance

        - shift the line closes to the radius (index 3) and shift according to the bisection method between 0 and the
        shortest distance

        :param radius: radius for which the shift is calculated
        :param area: area which is shifted in order to get in contact with a radius
        :param eps: bisection method accuracy
        :return: shift distance by which a trapezoidal area needs to be shifted in order to get in contact with radius
        (up to the eps accuracy)
        """
        # find point of crossing radial segment extension with radius
        # # lower line
        line_low = area.get_line(0)

        m_low, n_low = line_low.find_line_parameters()
        p_cross_low = Line.find_intersection_with_circle(radius, m_low, n_low)

        ## upper line
        line_up = area.get_line(2)
        m_up = m_low
        n_up = line_up.p2.y - m_up * line_up.p2.x
        p_cross_up = Line.find_intersection_with_circle(radius, m_up, n_up)

        if p_cross_low is None and p_cross_up is None:
            return 0.0

        # find maximum distance from the circle
        v_dir_low = p_cross_low - line_low.p1.copy()
        det_v_dir_l = v_dir_low.det()

        if abs(det_v_dir_l) <= 1e-12:
            return 0.0

        if p_cross_up is None:
            det_v_dir_u = 0
        else:
            v_dir_up = p_cross_up - line_up.p2.copy()
            det_v_dir_u = v_dir_up.det()

        # # take the maximum
        max_v_dir = max(det_v_dir_l, det_v_dir_u)

        # shift along this axis
        v_dir_l_unit = v_dir_low.unit()

        # lower point
        p_low = p_low_transl = area.get_line(0).p1.copy()
        # upper point
        line_left = area.get_line(3)
        det_line_left = line_left.det()
        alpha_left = Line.calculate_relative_alpha_angle(line_left)
        delta_alpha = 0 if alpha_left == 90 else 180
        p_up_transl = p_low.copy().translate(Point.of_polar(det_line_left, delta_alpha + alpha_left))

        start = 0
        end = max_v_dir
        shift = eps
        while (end - start) > eps:
            shift = (end + start) / 2
            p_low_transl = p_low.translate(v_dir_l_unit * shift)
            p_up_transl = p_low_transl.copy().translate(Point.of_polar(det_line_left, delta_alpha + alpha_left))

            if Line.find_interception_parameter(radius, Line.of_end_points(p_low_transl, p_up_transl)) is not None or \
                    Line.is_line_intercepting_radius(radius, Line.of_end_points(p_low_transl, p_up_transl)):
                end = shift
            else:
                start = shift

        if Line.find_interception_parameter(radius, Line.of_end_points(p_low_transl, p_up_transl)) is not None or \
            Line.is_line_intercepting_radius(radius, Line.of_end_points(p_low_transl, p_up_transl)):
            shift = start

        return shift

    @staticmethod
    def is_outside_of_first_quadrant(area: "Area", eps=1e-30) -> bool:
        """ Method checking whether an area is fully positioned in the first quadrant.

        :param area: an Area instance for which the check is performed
        :param eps: a machine precision for comparison of 0 value
        :return: True if an Area instance is in the first quadrant, False otherwise
        """
        is_x_nonegative = all([(line.p1.x >= -eps) and (line.p2.x >= -eps) for line in area.lines])
        is_y_nonegative = all([(line.p1.y >= -eps) and (line.p2.y >= -eps) for line in area.lines])
        return not (is_x_nonegative and is_y_nonegative)

    @staticmethod
    def are_areas_overlapping(area_a: "Area", area_b: "Area") -> True:
        p1 = Polygon([(area_a.get_line(i).p1.x, area_a.get_line(i).p1.y) for i in range(4)])
        p2 = Polygon([(area_b.get_line(i).p1.x, area_b.get_line(i).p1.y) for i in range(4)])
        return p1.intersects(p2)



from typing import List, Dict

import pandas as pd
from roxieapi.cadata.CableDatabase import CableDatabase

from magnumapi.geometry.GeometryBuilder import GeometryBuilder, SlottedGeometryBuilder
from magnumapi.geometry.Geometry import Geometry
import pymbse.commons.json_file as json_file
import roxieapi.tool_adapter.roxie_api as RoxieAPI


class GeometryFactory:
    """ GeometryFactory implements a factory design pattern and is used to produce:
    - rectangular geometry
    - absolute cos-theta geometry
    - relative cos-theta geometry
    - slotted absolute cos-theta geometry
    - slotted relative cos-theta geometry

    """

    @classmethod
    def init_with_json(cls, json_file_path: str, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a JSON file.

        :param json_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        json_content = json_file.read(json_file_path)
        return cls.init_with_dict(**json_content, cadata=cadata)

    @classmethod
    def init_with_dict(cls,
                       block_defs: List[Dict],
                       layer_defs: List[Dict],
                       cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a list of dictionaries with block definition.

        :param block_defs: a list of dictionaries with geometry definition (block definition)
        :param layer_defs: a list of dictionaries with layer definitions
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        return GeometryBuilder() \
            .with_block_defs(block_defs, cadata) \
            .with_layer_defs(layer_defs) \
            .build()

    @classmethod
    def init_slotted_with_json(cls, json_file_path: str, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a JSON file.

        :param json_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        json_content = json_file.read(json_file_path)
        return cls.init_slotted_with_dict(**json_content, cadata=cadata)

    @classmethod
    def init_slotted_with_dict(cls,
                               block_defs: List[Dict],
                               layer_defs: List[Dict],
                               cadata: CableDatabase,
                               r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a list of dictionaries with block definition.

        :param block_defs: a list of dictionaries with geometry definition (block definition)
        :param layer_defs: a list of dictionaries with layer definitions
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        return SlottedGeometryBuilder() \
            .with_block_defs(block_defs, cadata) \
            .with_layer_defs(layer_defs) \
            .with_r_aperture(r_aperture) \
            .build()

    @classmethod
    def init_with_data(cls, data_file_path: str, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a DATA ROXIE file.

        :param data_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        block_df = RoxieAPI.read_bottom_header_table(data_file_path, keyword='BLOCK')
        layer_df = RoxieAPI.read_nested_bottom_header_table(data_file_path, keyword='LAYER')
        return cls.init_with_df(block_df, layer_df, cadata)

    @classmethod
    def init_with_csv(cls,
                      block_csv_file_path: str,
                      layer_csv_file_path: str,
                      cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a CSV file.

        :param block_csv_file_path: a path to a csv file with block definitions
        :param layer_csv_file_path: a path to a csv file with layer definitions
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        block_df = pd.read_csv(block_csv_file_path, index_col=0)
        layer_df = pd.read_csv(layer_csv_file_path, index_col=0)
        return cls.init_with_df(block_df, layer_df, cadata)

    @classmethod
    def init_with_df(cls, block_df: pd.DataFrame, layer_df, cadata: CableDatabase) -> Geometry:
        """ Class method initializing a Geometry instance from a dataframe with block definition.

        :param block_df: a dataframe with geometry definition (block definition)
        :param cadata: a CableDatabase instance
        :return: initialized geometry instance
        """
        return GeometryBuilder() \
            .with_block_df(block_df, cadata) \
            .with_layer_df(layer_df) \
            .build()

    @classmethod
    def init_slotted_with_data(cls, data_file_path: str, cadata: CableDatabase, r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a DATA ROXIE file.

        :param data_file_path: a path to a json file
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        block_df = RoxieAPI.read_bottom_header_table(data_file_path, keyword='BLOCK')
        layer_df = RoxieAPI.read_nested_bottom_header_table(data_file_path, keyword='LAYER')
        return cls.init_slotted_with_df(block_df, layer_df, cadata, r_aperture)

    @classmethod
    def init_with_csv_slotted(cls,
                      block_csv_file_path: str,
                      layer_csv_file_path: str,
                      cadata: CableDatabase,
                      r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a CSV file.

        :param block_csv_file_path: a path to a csv file with block definitions
        :param layer_csv_file_path: a path to a csv file with layer definitions
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        block_df = pd.read_csv(block_csv_file_path, index_col=0)
        layer_df = pd.read_csv(layer_csv_file_path, index_col=0)
        return cls.init_slotted_with_df(block_df, layer_df, cadata, r_aperture)

    @classmethod
    def init_slotted_with_df(cls,
                             block_df: pd.DataFrame,
                             layer_df: pd.DataFrame,
                             cadata: CableDatabase,
                             r_aperture: float) -> Geometry:
        """ Class method initializing a Geometry instance from a dataframe with block definition.

        :param block_df: a dataframe with geometry definition (block definition)
        :param cadata: a CableDatabase instance
        :param r_aperture: aperture radius in mm
        :return: initialized geometry instance
        """
        return SlottedGeometryBuilder() \
            .with_block_df(block_df, cadata) \
            .with_layer_df(layer_df) \
            .with_r_aperture(r_aperture) \
            .build()



from dataclasses import dataclass

from magnumapi.geometry.definitions.BlockDefinition import BlockDefinition


@dataclass
class RectangularBlockDefinition(BlockDefinition):
    """Class for a rectangular block definition.

    Attributes:
       x (float): The x-coordinate of lower left block corner (assuming alpha equal to 0).
       y (float): The y-coordinate of lower left block corner (assuming alpha equal to 0).
       alpha (float): The inclination angle in degrees.
    """
    x: float
    y: float
    alpha: float



from dataclasses import dataclass


@dataclass
class BlockDefinition:
    """Class for block definition.

   Attributes:
       no (int): The row index of a block definition
       type (int): The block type (e.g., 1 - cos-theta, 2 - rectangular).
       nco (int): The number of conductors.
       current (float): The current in a block.
       condname (str): The name of a conductor in the cable database.
       n1 (int): The number of strands in the narrow direction.
       n2 (int): The number of strands in the wide direction.
       imag (int): 1: Block imaged at x-axis; 0: No action. (not with "Symmetric Coil" -option).
       turn (float): Block turned by angle. (not with "Symmetric Coil" -option).
    """
    no: int
    type: int
    nco: int
    current: float
    condname: str
    n1: int
    n2: int
    imag: int
    turn: float


from dataclasses import dataclass

from magnumapi.geometry.definitions.CosThetaBlockDefinition import CosThetaBlockDefinition


@dataclass
class RelativeCosThetaBlockDefinition(CosThetaBlockDefinition):
    """Class for relative cos-theta block definition.

    Attributes:
       phi_r (float): The positioning angle in degrees.
       alpha_r (float): The inclination angle in degrees.
    """
    phi_r: float
    alpha_r: float



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magnumapi',
 'magnumapi.geometry',
 'magnumapi.geometry.blocks',
 'magnumapi.geometry.definitions',
 'magnumapi.geometry.primitives',
 'magnumapi.matpro',
 'magnumapi.plotting',
 'magnumapi.tool_adapters',
 'magnumapi.tool_adapters.ansys']

package_data = \
{'': ['*']}

install_requires = \
['Shapely>=1.8.2,<2.0.0',
 'ansys-mapdl-core>=0.61.6,<0.62.0',
 'ipyaggrid>=0.2.2,<0.3.0',
 'matplotlib>=3.5.2,<4.0.0',
 'numpy>=1.22.4,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'pymbse-commons>=0.0.3,<0.0.4',
 'roxie-api>=0.0.4,<0.0.5']

setup_kwargs = {
    'name': 'magnum-api',
    'version': '0.1.49',
    'description': '',
    'long_description': '# Project Overview\nThe CHART MagNum (Magnum Numerics) aims at introducing advanced design techniques in order to support the design process of superconducting accelerator magnet.\nThe project responds to the following strategic goals:\n- Sustainability\n\n  Ensure that outstanding modeling work will have an impact on present or future designs. This also require that models have clearly defined scope and range of applicability.\n- Traceability\n\n  Ensure that the modeler is able to trace back the input parameters, code and script versions, etc., that have been used to produce a particular plot in a ppt or pdf.\n  Traceability is even more important in multi-scale and multi-model analysis.\n- Repeatability\n\n  Ensure that results presented at, e.g., a Conceptual Design Review can be reproduced at any later time.\n  Ensure that as-built models can be re-run at any moment during a potentially decades-long project life cycle.\n- Flexibility\n\n  Allow for different labs and collaborators to have/prefer different licenses.\n  Enable researchers to implement innovative ideas while building upon existing best practices, but without having to solve legacy issues.\n- Usability\n\n  Encapsulate the increased flexibility behind easy-to-use UIs for the standard design work.\n\nThe project implements a number of concepts introduced by the MBSE (Model-Based System Engineering) methdology.\nIn particular, MBSE shifts the focus from documents to models as primary means of communication in complex system design projects.\n\n# Project Architecture\nThe project is composed of two parts: notebooks and a library with an API (Application Programming Interface).\nA notebook combines documentation with source code and results of its execution. As such, it encodes a design step that involves model construction from a parametrized input, its execution and extraction of the figures of merit.\nIn order to keep the notebooks lean and linear, the model control logic is extracted to the API.\nAn important pillar of the framework is automation of the modelling process; the automation is achieved with GitLab CI pipelines. To this end, a set of notebooks representing a design process are concatenated into a design workflow.\nThe workflow is executed for a given model input. The input contains generic part (e.g., material properties, geometry, cable parameters) and model-specific part (e.g., electromagnetic boundary condidtions, definition of mechanical contacts).\nAs a result, a workflow is immediatelly suited for design optimization, co-simulation, parametric studies, etc.\n\n# User Guide\nIn the following we discuss the steps needed to obtain and further develop the API. In addition,\nthe main modules of `magnum-api` are presented:\n- Geometry\n- Tool Adapters\n- Optimization\n- Material Properties\n- Plotting\n\n## Getting Started\nThe project is developed with Python 3.6+ and was tested on Windows and Linux machines.\nThe python project dependencies are stored in `requirements.txt` file and need to be installed prior to API execution.\nIn order to execute tool adapters for ROXIE and ANSYS, both need to be installed. The API supports ROXIE 10.3 and ANSYS 2020R1.\nThe code development can be carried out with any IDE (Integrated Development Environment) and the suggested one is PyCharm.\nPlease consult `CONTRIBUTING.md` file for contribution guidelines.\n\n## Geometry\nThe geometry module provides classes to handle coil geometry. So far, the 2D cos-theta geometry is implemented.\nWe follow the ROXIE block definition and reuse the cable data concept. The block definition supports absolute and relative angle definition. The geometry input can be expressed as a list of dictionaries and a pandas DataFrame either as objects or, respectively as json and csv files.\n\n- geometry creation with json file for absolute angle definition\n\n```python\nfrom magnumapi.cadata.CableDatabase import CableDatabase\nfrom magnumapi.geometry.CosThetaGeometry import CosThetaGeometry\n\njson_path = \'input/16T_abs.json\'\ncadata_path = \'input/roxieold_2.cadata\'\ncadata = CableDatabase(cadata_path)\n\ngeometry = CosThetaGeometry.with_roxie_absolute_angle_definition_json(json_path, cadata)\ngeometry.build_blocks()\ngeometry.plotly_geometry_blocks()\n```\n\n- geometry creation with json file for relative angle definition\n\n```python\nfrom magnumapi.cadata.CableDatabase import CableDatabase\nfrom magnumapi.geometry.CosThetaGeometry import CosThetaGeometry\n\njson_path = \'input/16T_rel.json\'\ncadata_path = \'input/roxieold_2.cadata\'\ncadata = CableDatabase(cadata_path)\n\ngeometry = CosThetaGeometry.with_roxie_relative_angle_definition_json(json_path, cadata)\ngeometry.build_blocks()\ngeometry.plotly_geometry_blocks()\n```\n\nThe remaining methods are presented in the UML diagram below.\n<img src="figures/CosThetaGeometryUMLClassDiagram.PNG">\n\n## Tool Adapters\nA tool adapter provides methods to control a simulation tool. So far, we support ROXIE and ANSYS as show in the diagram below. \n<img src="figures/CosThetaGeometryUMLClassDiagram.PNG">\n\n## Optimization\n\nWe implement a genetic optimization algorithm that operates with modelling workflows implemented as notebooks.\nThe optimization algorithm works as follows:\n1. Load design variables table with allowed range of values, its type, and number of bits per variable\n2. Load optimization config\n3. Load an input model parameter that is updated during the optimization process. Each notebook relies on this input.\n4. Randomly initialize starting population of individuals\n5. For each generation:\n  - Evaluate fitness function for each notebook (in case of errors, a penalty value is applied)\n  - Perform tournament selection for mating (if elitism is considered)\n  - Execute crossover operator\n  - Execute mutation operator\n\n### Design Variables Table\nBelow we present a sample design variables table. The meaning of columns is\n- xl: lower value of a variable\n- xu: upper value of a variable\n- xs: starting value of a variable\n- variable: variable name\n- variable_type: numeric variable type\n- bits: bits per variable\n- bcs: block index\n\n| xl        | xu           | xs  | variable | variable_type | bits | bcs |\n| ------------- |-------------| -----| ---- | ------------- | ---- | --- |\n| 0 | 3.0 | 10.0 | 3 | phi_r | float | 6 | 2 |\n| 1 | 3.0 | 10.0 | 3 | phi_r | float | 6 | 3 |\n| 2 | 3.0 | 10.0 | 3 | phi_r | float | 6 | 4 |\n| 3 | 3.0 | 10.0 | 3 | phi_r | float | 6 | 6 |\n| 4 | 3.0 | 10.0 | 3 | phi_r | float | 6 | 7 |\n\n### Optimization Config\nOptimization algorithm is parametrized with a config. The config is represented as a json file.\nA sample config is shown below. The meaning of keys is:\n- input_folder: an absolute path to an input folder\n- output_folder: an absolute path to an output folder\n- notebooks: a list of notebook configs\n  - notebook_folder: relative notebook folder path\n  - notebook_name: full notebook name\n  - input_parameters: dictionary of parameters; key is a name for the target notebook and the value is the name from a source notebook\n  - output_parameters: list of output parameters\n  - input_artefacts: dictionary of parameters; key is a relative artefact path for the target notebook and the value is the relative artefact path from a source notebook\n  - output_artefacts: list of output artefacts\n\n```json\n{\n  "input_folder": "/home/mmaciejewski/gitlab/magnum-nb/",\n  "output_folder": "/home/mmaciejewski/gitlab/magnum-nb/output/",\n  "notebooks": [\n    {\n      "notebook_folder": "geometry",\n      "notebook_name": "Geometry.ipynb",\n      "input_parameters": {},\n      "output_parameters": [],\n      "input_artefacts": {},\n      "output_artefacts": []\n    },...\n  ]\n}\n```\n\n## Material Properties\n\nWe implemented the following material properties:\n\n- resistivities:\n  - calc_rho_cu_nist\n- heat capacities:\n  - calc_cv_cu_nist\n  - calc_cv_nb3sn_nist\n- critical current fits:\n  - calc_jc_nbti_bottura\n  - calc_jc_nb3sn_bordini\n  - calc_jc_nb3sn_summers\n  - calc_jc_nb3sn_summers_orig\n\n## Plotting\nPlotting module contains generic plotting functions. \n',
    'author': 'mmaciejewski',
    'author_email': 'michal.maciejewski@ief.ee.ethz.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.cern.ch/chart-magnum/magnum-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
from typing import List

import pandas as pd
import matplotlib.pyplot as plt

from roxieapi.cadata.CableDefinition import CableDefinition
from roxieapi.cadata.ConductorDefinition import ConductorDefinition
from roxieapi.cadata.InsulationDefinition import InsulationDefinition
from roxieapi.cadata.StrandDefinition import StrandDefinition

from magnumapi.geometry.definitions.BlockDefinition import BlockDefinition
from magnumapi.geometry.primitives.Area import Area


class Block(ABC):
    """ An Block class implementing a container with block definitions from cadata input.
    In addition, the class stores initialized coordinates of all conductors per block.
    This is a base class extended by implementations of different types (cos-theta, rectangular, etc.).
    """

    def __init__(self,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        """ Method constructing an instance of a Block class

        :param cable_def: a cable definition from cadata
        :param insul_def: an insulation definition from cadata
        :param strand_def: a strand definition from cadata
        :param conductor_def: a conductor definition from cadata
        """
        self.cable_def = cable_def
        self.insul_def = insul_def
        self.strand_def = strand_def
        self.conductor_def = conductor_def
        self.block_def: BlockDefinition = None
        self.areas: List[Area] = []

    def compute_surface(self):
        """ Method computing surface of a block as S = nco * ns * S_strand,
        where S_strand = pi * d_strand^2 / 4

        :return: a surface of a block in mm^2
        """
        return self.block_def.nco * self.cable_def.n_s * self.strand_def.compute_surface()

    def compute_surface_cu(self):
        """ Method computing copper surface of a block as S = nco * ns * S_strand_cu,
        where S_strand_cu = f_cu * pi * d_strand^2 / 4

        :return: a copper surface of a block in mm^2
        """
        return self.block_def.nco * self.cable_def.n_s * self.strand_def.compute_surface_cu()

    def compute_surface_nocu(self):
        """ Method computing copper surface of a block as S = nco * ns * S_strand_nocu,
        where S_strand_nocu = f_nocu * pi * d_strand^2 / 4

        :return: a non-copper surface of a block in mm^2
        """
        return self.block_def.nco * self.cable_def.n_s * self.strand_def.compute_surface_nocu()

    @abstractmethod
    def plot_block(self, ax: plt.Axes) -> None:
        """ Abstract method for plotting an insulated block

        :param ax: a matplotlib axis on which a block will be plotted
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def plot_bare_block(self, ax: plt.Axes) -> None:
        """ Abstract method for plotting a bare block

        :param ax: a matplotlib axis on which a bare block will be plotted
        """
        raise NotImplementedError('This method is not implemented for this class')

    def empty_areas(self) -> None:
        """ Method setting areas to an empty array

        """
        self.areas = []

    def get_bare_areas(self) -> List[Area]:
        """ Method returning bare areas, i.e., areas without insulation

        :return: a list of areas without insulation of a given block
        """
        return [self.get_bare_area(area) for area in self.areas]

    def get_bare_area(self, area_ins: Area) -> Area:
        """ Abstract method for returning a bare area

        :param area_ins: an insulated area
        :return: a bare area
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def build_block(self) -> None:
        """ Abstract method for building a block

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def to_block_df(self) -> pd.DataFrame:
        """ Abstract method for converting a block definition into a ROXIE-compatible dataframe

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def to_abs_dict(self):
        """ Abstract method for returning the absolute dictionary for block definition

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def to_rel_dict(self, alpha_ref: float, phi_ref: float):
        """ Abstract method for returning the absolute dictionary for block definition

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def homogenize(self):
        """ Abstract method for homogenizing a block

        """
        raise NotImplementedError('This method is not implemented for this class')

    def to_df(self) -> pd.DataFrame:
        """ Method concatenating rows representing coordinates of each area into a dataframe

        :return: a dataframe with coordinate information of each area of a block instance
        """
        return pd.concat([area.to_df() for area in self.areas], axis=0)

    def is_outside_of_first_quadrant(self, eps=1e-30) -> True:
        """ Method checking whether at least one area of a given block are outside of the first quadrant

        :param eps: a machine precision for comparison of 0 value
        :return: True if at least one area of a given block is outside of the first quadrant,
        False otherwise, i.e., if all areas are within the first quadrant.
        """
        if not self.areas:
            return False

        is_inside_first_quadrant = False
        for area in self.areas:
            is_inside_first_quadrant |= Area.is_outside_of_first_quadrant(area, eps=eps)

        return is_inside_first_quadrant



import numpy as np


def calc_jc_nbti_bottura(T, B, Tc0=9.2, Bc20=14.5, Jc_ref=3.0e9, C0=27.04, alpha=0.57, beta=0.8477, gamma=2.23234):
    # Direction of the magnetic field is not important
    B = abs(B)

    # Calculate critical magnetic field
    Bc2 = Bc20 * (1 - np.power(T / Tc0, 1.7))

    # Fix too small values
    Bc2 = np.maximum(Bc2, 0.001)

    # Calculate critical current density
    return Jc_ref * C0 * np.power(B, alpha - 1) / np.power(Bc2, alpha) * np.power(1 - B / Bc2, beta) \
           * np.power(1 - np.power(T / Tc0, 1.7), gamma)


def calc_jc_nb3sn_bordini(T, B, Jc0=267845e6, Tc0=16.3, Bc20=26.45, alpha=1.0):
    # Direction of the magnetic field is not important
    B = abs(B)

    # Very small magnetic field causes numerical problems
    B = np.maximum(B, 0.001)

    # Ratio of temperature to critical temperature
    # Avoid values higher than 1
    f_T_T0 = np.minimum(T / Tc0, 1.0)

    # Calculate critical magnetic field
    Bc2 = Bc20 * (1 - np.power(f_T_T0, 1.52))

    # Ratio of magnetic field to critical magnetic field
    f_B_Bc2 = B / Bc2

    # Avoid values higher than 1
    f_B_Bc2 = np.minimum(f_B_Bc2, 1.0)

    C = Jc0 * np.power(1 - np.power(f_T_T0, 1.52), alpha) * np.power(1 - np.power(f_T_T0, 2.0), alpha)
    return C / B * np.power(f_B_Bc2, 0.5) * np.power(1 - f_B_Bc2, 2.0)  # A / m^2


def calc_jc_nb3sn_summers(T, B, Tc0=18.0, Bc20=28.0, Jc0=3.85e10):
    # Direction of the magnetic field is not important
    B = abs(B)

    # Very small magnetic field causes numerical problems
    B = np.maximum(B, 0.001)

    # Very small temperature causes numerical problems
    T = np.maximum(T, 0.001)

    # Avoid values higher than 1
    f_T_T0 = np.minimum(T / Tc0, 1.0)
    Bc2 = Bc20 * (1 - np.power(f_T_T0, 2)) * (1 - 0.31 * np.power(f_T_T0, 2) * (1 - 1.77 * np.log(f_T_T0)))
    f_B_Bc2 = B / Bc2

    # Avoid values higher than 1
    f_B_Bc2 = np.minimum(f_B_Bc2, 1.0)

    return Jc0 / np.sqrt(B) * np.power(1 - f_B_Bc2, 2) * np.power((1 - np.power(f_T_T0, 2)), 2)


def calc_jc_nb3sn_summers_orig(T, B, Tc0=18.0, Bc20=28.0, Jc0=3.85e10):
    # Direction of the magnetic field is not important
    B = abs(B)

    # Very small magnetic field causes numerical problems
    B = np.maximum(B, 0.001)

    # Very small temperature causes numerical problems
    T = np.maximum(T, 0.001)

    # Avoid values higher than 1
    f_T_T0 = np.minimum(T / Tc0, 1.0)
    Bc2 = Bc20 * (1 - np.power(f_T_T0, 2)) * (1 - 0.31 * np.power(f_T_T0, 2) * (1 - 1.77 * np.log(f_T_T0)))
    f_B_Bc2 = B / Bc2

    # Avoid values higher than 1
    f_B_Bc2 = np.minimum(f_B_Bc2, 1.0)

    return Jc0 / np.sqrt(B) * np.power(1 - f_B_Bc2, 2) * 1 / np.sqrt(f_B_Bc2) * np.power((1 - np.power(f_T_T0, 2)), 2)





import math
import warnings
from dataclasses import dataclass
from typing import Tuple, Union

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Point import Point


@dataclass(frozen=True)
class Line:
    """ A Line class implementing a 2D line made of two end points along with basic geometric transformations

   Attributes:
       p1 (Point): start point of a line
       p2 (Point): end point of a line
    """
    p1: Point
    p2: Point

    @staticmethod
    def of_end_points(p1: Point, p2: Point) -> "Line":
        """ Static method initiating a line from two end points

        :param p1: start point of a line
        :param p2: end point of a line
        :return: Initialized line object with copies of both lines
        """
        return Line(p1=p1.copy(), p2=p2.copy())

    def translate(self, vector: Point) -> "Line":
        """ Method translating line by an input vector

        :param vector: vector by which the line is translated
        :return: translated line
        """
        return Line.of_end_points(self.p1.translate(vector), self.p2.translate(vector))

    def rotate(self, angle: float) -> "Line":
        """ Method rotating line by a given angle w.r.t. origin of the coordinate system (0, 0)

        :param angle: rotation angle in degrees
        :return: rotated line
        """
        return Line.of_end_points(self.p1.rotate(angle), self.p2.rotate(angle))

    def unit(self) -> Point:
        """ Method calculating a unit vector along the line

        :return: Unit vector along the line with start at the origin of the coordinate system
        """
        vector = self.p1.copy() - self.p2.copy()
        return vector.copy().unit()

    def is_constant(self) -> bool:
        """ Method checking whether the line is constant (y values are equal)

        :return: True if y values are equal, otherwise False
        """
        return self.p1.y == self.p2.y

    def find_line_parameters(self) -> Tuple[float, float]:
        """ Method finding line parametrization as y = m*x + n
        p1.y = m*p1.x + n
        p2.y = m*p2.x + n
        subtracting both sides
        p1.y - p2.y = m*p1.x - m*p2.x
        m = (p1.y - p2.y) / (p1.x - p2.x)
        n = p1.y - m*p1.x

        :return: a Tuple with the first element being m and the second being n
        """
        m = (self.p1.y - self.p2.y) / (self.p1.x - self.p2.x)
        n = self.p1.y - m * self.p1.x
        return m, n

    def __str__(self) -> str:
        """ Method returning a string representation of a Line for a user

        :return: string representation of a Line instance
        """
        return "%s -> %s" % (self.p1, self.p2)

    def to_df(self) -> pd.DataFrame:
        """ Method converting a line object into a dataframe with end point coordinates

        :return: a DataFrame with x1, y1, x2, y2 columns
        """
        return pd.DataFrame(columns=['x1', 'y1', 'x2', 'y2'],
                            data=[[self.p1.x, self.p1.y, self.p2.x, self.p2.y]],
                            index=[0])

    @staticmethod
    def find_intersection_with_circle(r: float, m: float, n: float) -> Union[None, Point]:
        """ Method finding intersection of a circle given by radius and line given by y = m*x + n parametrization
        (x-x_0)^2 + (y-y_0)^2 = r^2
        y = m*x+n
        for simplicity, x_0 = 0, y_0 = 0
        x^2 + y^2 = r^2
        x^2 + (mx+n)^2 = r^2
        x^2 + m^2*x^2 + 2*m*n*x + n^2 - r^2 = 0
        (1 + m^2) * x^2 + 2*m*n*x + n^2 - r^2 = 0
        Let, a = 1 + m^2, b = 2*m*n, c = n^2 - r^2
        delta = b^2 - 4*a*c
        if delta < 0 - a real valued solution does not exist
        otherwise, there are two solution exist, one with negative x and the other with positive,
        we take positive one assuming that we are in the first quadrant
        x = (-b + sqrt(delta)) / 2a
        y = m*x + n
        :param r: radius of a circle
        :param m: m line coefficient
        :param n: n line coefficient
        :return: If there is an intersection, then Point is returned, otherwise None
        """

        a = 1 + m ** 2
        b = 2 * m * n
        c = n ** 2 - r ** 2

        delta = b ** 2 - 4 * a * c
        if delta < 0:
            return None
        else:
            x = (-b + delta ** 0.5) / (2 * a)
            y = m * x + n
            return Point.of_cartesian(x, y)

    def copy(self) -> "Line":
        """ Method returning a copy of a current Line instance

        :return: a copy of a Line instance
        """
        return Line.of_end_points(p1=self.p1.copy(), p2=self.p2.copy())

    def det(self) -> float:
        """ Method calculating line length

        :return: lenght of the line
        """
        return (self.p2 - self.p1).det()

    @staticmethod
    def calc_arc_length_between_two_lines(radius: float, line_low: "Line", line_up: "Line") -> float:
        """ Static method calculating arc lenght between two lines. The method is used to compute wedge length at
        its tip between two blocks. The method uses the law of cosines in order to compute the length. It is assumed
        that both lines are defined with the same unit.

        :param radius: radius with which the lines are intercepting
        :param line_low: lower line
        :param line_up: upper line
        :return: arc length in units determined from the line definitions
        """
        # find crossing points of both lines with the radius
        m_low, n_low = line_low.find_line_parameters()
        p_low = Line.find_intersection_with_circle(radius, m_low, n_low)
        m_up, n_up = line_up.find_line_parameters()
        p_up = Line.find_intersection_with_circle(radius, m_up, n_up)
        # calculate d
        d = (p_low - p_up).det()
        # calculate theta from the law of cosines
        theta = math.acos(1 - d ** 2 / (2 * radius ** 2))
        l_arc = radius * theta
        return l_arc

    @staticmethod
    def calculate_point_orientation_wrt_line(line: "Line", point: Point, eps=1e-12) -> int:
        """ Method calculating orientation of a point w.r.t a line
        The orientation is obtained from the sign of a matrix (http://www.cs.cmu.edu/~quake/robust.html)
        - if the sign is positive, then the point is on the left side
        - if the sign is negative, then the point is on the right sied
        - if the sign equal to 0 (with eps tolerance), then it is assumed that the point is on the line

        :param line: reference line
        :param point: point for which the orientation is checked
        :return: -1 (on the right), +1 (on the left), 0 (on the line)
        """
        matrix = np.array([[line.p1.x - point.x, line.p1.y - point.y],
                           [line.p2.x - point.x, line.p2.y - point.y]])
        det = np.linalg.det(matrix)
        if abs(det) < eps:
            return 0
        else:
            return np.sign(det)

    @staticmethod
    def is_line_intercepting_radius(radius: float, line: "Line") -> bool:
        """ Method calculating whether a line is intercepting a radius.

        :param radius: radius to check
        :param line: line to calculate the interception
        :return: True if at least one point is within the radius, otherwise False
        """
        return (line.p1.get_r() < radius) or (line.p2.get_r() < radius)

    @staticmethod
    def find_interception_parameter(r: float, line: "Line") -> Union[None, float]:
        """ Method finding an interception parameter of a line with a radius
        Assuming that the center of the circle is at 0,0
        The segment S form p1 to p2 is parameterised as
        S = p1 + t(p2-p1) = p1 + t*v for v = p2-p1, and t = [0, 1]
        The segment intercepts the circle when
        |S - q| = r, with our assumption:
        |S| = r, |p1 + t*v| = r
        Squaring both sides we obtain:
        |p1 + t*v|^2 = r^2
        (p1 + t*v)^2 = r^2
        p1^2 + 2*t*v*p1 + (t*v)^2 - r^2 = 0
        (t*v)^2 + 2*t*v + p1^2 - r^2 = 0
        then, a = v^2, b = 2*v, c = p1^2 - r^2
        Then, delta is equal to b**2 - 4*a*c . For negative delta, the line does not intercept radius. Otherwise,
        t1 and t2 can be computed. Only if one of the solutions is in [0, 1] range, then the line segment intercepts a
        radius. Values outside of this range indicate that the segment would need to be extended in order to cross a
        radius.

        :param r: radius with center at 0, 0
        :param line: line for which the interception parameter is calculated
        :return: None if there is no intersection
        """

        v = line.p2 - line.p1
        a = v.dot(v)
        b = 2 * v.dot(line.p1)
        c = line.p1.dot(line.p1) - r ** 2

        delta = b ** 2 - 4 * a * c

        if delta < 0:
            return None

        t1 = (-b + delta ** (0.5)) / (2 * a)
        t2 = (-b - delta ** (0.5)) / (2 * a)

        if 0 <= t1 <= 1:
            return t1
        elif 0 <= t2 <= 1:
            return t2
        else:
            return None

    def plot(self, ax: plt.Axes, color="black") -> None:
        """ Method plotting an Area instance on an input axis

        :param ax: input axis for plot
        :param color: color for line plotting, black by default
        """
        ax.plot([self.p1.x, self.p2.x], [self.p1.y, self.p2.y], color=color)

    @staticmethod
    def calculate_positioning_angle(line_ref: "Line", radius: float) -> float:
        """ Method calculating the positioning angle from the point of line crossing a radius

        :param line_ref: reference line for which the positioning angle is calculated
        :param radius: radius for which the positioning angle is calculated
        :return: positioning angle in degrees
        """
        if line_ref.is_constant():
            y = line_ref.p1.y
            p_cross = Point.of_cartesian(x=math.sqrt(radius ** 2 - y ** 2), y=y)
        else:
            m, n = line_ref.find_line_parameters()
            p_cross = Line.find_intersection_with_circle(radius, m, n)

        if p_cross is None:
            warning_msg = """A line used to calculate the positioning angle %s does not intercept the radius %f.
            Instead, returned the phi angle of the first point of the reference line""" % (line_ref, radius)
            warnings.warn(warning_msg)
            return line_ref.p1.get_phi()

        return p_cross.get_phi()

    @staticmethod
    def calculate_relative_alpha_angle(line_ref: "Line") -> float:
        """ Static method calculating the inclination angle of a line, i.e., its slope angle in degrees.
        There are three cases:
        - 0 degrees if the line is constant,
        - 90 degrees if the x coordinates are the same,
        - arcus tangent of the slope coefficient
        The method is used to calculate block inclination angle

        :param line_ref: a Line instance for which the angle is calculated
        :return: slope angle in degrees
        """
        if line_ref.is_constant():
            alpha_ref = 0
        elif line_ref.p1.x == line_ref.p2.x and (line_ref.p1.y < line_ref.p2.y):
            alpha_ref = 90
        elif line_ref.p1.x == line_ref.p2.x and (line_ref.p1.y > line_ref.p2.y):
            alpha_ref = 270
        else:
            m, _ = line_ref.find_line_parameters()
            alpha_ref = math.degrees(math.atan(m))

        return alpha_ref
